@@ -12,6 +12,7 @@ cases, users will not construct these objects directly but rather through use of
 class.
 """
 
+import re
 from .algebra import Atom
 from functools import reduce
 from enum import Enum
@@ -20,6 +21,9 @@ from enum import Enum
 __author__ = 'Evan W. Patton <ewpatton@mit.edu>'
 
 
+NAMESPACE = re.compile('\\{[^}]+}')
+
+
 def _html(tag):
     """
     Prefixes the given ``tag`` with the XHTML prefix. Block files saved with newer versions of App Inventor will have
@@ -35,7 +39,7 @@ def _html(tag):
 # noinspection PyShadowingBuiltins
 class Block(object):
     _CATEGORIES = {'color', 'component', 'controls', 'global', 'lexical', 'lists', 'local', 'logic', 'math',
-                   'obfuscated', 'dictionaries', 'procedures', 'text'}
+                   'obfuscated', 'dictionaries', 'procedures', 'text', 'pair', 'helpers'}
     _CATEGORY_MAP = {
         'color': 'Colors',
         'component': 'Components',
@@ -49,8 +53,10 @@ class Block(object):
         'obfuscated': 'Text',
         'obsfucated': 'Text',
         'dictionaries': 'Dictionaries',
+        'pair': 'Dictionaries',
         'procedures': 'Procedures',
-        'text': 'Text'
+        'text': 'Text',
+        'helpers': 'Helpers'
     }
     _ID_COUNT = 0
 
@@ -89,23 +95,42 @@ class Block(object):
         if id is None:
             id = Block._ID_COUNT
             Block._ID_COUNT += 1
+        if type == 'procedures_do_then_return':
+            type = 'controls_do_then_return'
+        elif type == 'procedure_lexical_variable_get':
+            type = 'lexical_variable_get'
+        elif type == 'for_lexical_variable_get':
+            type = 'lexical_variable_get'
         type_parts = type.split('_')
-        if type_parts[0] not in Block._CATEGORIES and lang_ver < 17:
+        extra_mutations = {}
+        if type_parts[0] not in Block._CATEGORIES and (lang_ver is None or lang_ver < 17):
             # likely old-format blocks code with component names in block types
-            if type_parts[0] in screen.components:
+            component = screen.components(lambda x: x.name == type_parts[0])
+            if len(component) > 0:
+                extra_mutations['instance_name'] = type_parts[0]
+                extra_mutations['component_type'] = component[0].type.name
                 if type_parts[1] == 'setproperty':
                     # Old-style property setter
                     type = 'component_set_get'
+                    extra_mutations['property_name'] = type_parts[1]
+                    extra_mutations['set_or_get'] = 'set'
                 elif type_parts[1] == 'getproperty':
                     # Old-style property getter
                     type = 'component_set_get'
+                    extra_mutations['property_name'] = type_parts[1]
+                    extra_mutations['set_or_get'] = 'get'
                 elif len(xml) >= 2 and xml[1].tag == 'statement' and 'name' in xml[1].attrib and \
                         xml[1].attrib['name'] == 'DO':
                     # Old-style event handler
                     type = 'component_event'
+                    extra_mutations['event_name'] = type_parts[1]
                 else:
                     # Old-style method call
                     type = 'component_method'
+                    extra_mutations['method_name'] = type_parts[1]
+            elif type_parts[0] == 'obsufcated':
+                type_parts = ['obfuscated', 'text']
+                type = '_'.join(type_parts)
             else:
                 raise RuntimeError('Unknown block type: %s' % type_parts[0])
         block = Block(id, type)
@@ -126,9 +151,16 @@ class Block(object):
         for child in xml:
             if child.tag == 'mutation' or child.tag == _html('mutation'):
                 block.mutation = dict(child.attrib)
+                block.mutation.update(extra_mutations)
+                extra_mutations = {}
                 if type.startswith('component_') and ('is_generic' not in block.mutation or
                                                       block.mutation['is_generic'] == 'false'):
                     block.component = screen.components[block.mutation['instance_name']]
+                for grandchild in child:
+                    tag = '.' + NAMESPACE.sub('', grandchild.tag)
+                    if tag not in block.mutation:
+                        block.mutation[tag] = []
+                    block.mutation[tag].append(dict(grandchild.attrib))
             elif child.tag == 'comment' or child.tag == _html('comment'):
                 block.comment = child.text
             elif child.tag in {'field', 'title', _html('field'), _html('title')}:
@@ -150,6 +182,8 @@ class Block(object):
                 child_block = Block.from_xml(screen, child[0], lang_ver, siblings=siblings, parent=block,
                                              connection_type='next')
                 block.next = child_block
+        if len(extra_mutations) > 0:  # mutations were not consumed, so we're missing a <mutation> tag
+            block.mutation = extra_mutations
         return block
 
     @property
@@ -179,6 +213,16 @@ class Block(object):
         return reduce(list.__add__, iter(self.values.values()), []) + \
                reduce(list.__add__, iter(self.statements.values()), [])
 
+    def has_mutation(self, name):
+        """
+        Tests whether the block has a given mutation.
+
+        :param str name: the name of the mutation to test
+        :return: True if the mutation is present on the block, otherwise False
+        :rtype: bool
+        """
+        return False if self.mutation is None else (name in self.mutation)
+
     def __repr__(self):
         return '%s(%r, %r)' % (self.__class__.__name__, self.id, self.type)
 
